from ._frame import Command
from ._base import ResponseException, Functionality, Irq
try:
  from enum import Enum
except ImportError:
  from enum34 import Enum


class DigitalInputs(Functionality):
    """Attributes and methods needed for operating the digital inputs channels.

    Args:
        i2c_hat (:obj:`raspihats.i2c_hats._base.I2CHat`): I2CHat instance
        labels (:obj:`list` of :obj:`str`): Labels of digital input channels

    Attributes:
        channels (:obj:`list` of :obj:`bool`): List like object, provides access to digital inputs channels.
        r_counters (:obj:`list` of :obj:`int`): List like object, provides access to raising edge digital input counters.
        f_counters (:obj:`list` of :obj:`int`): List like object, provides access to falling edge digital input counters.
    """

    def __init__(self, i2c_hat, labels):
        Functionality.__init__(self, i2c_hat, labels)
        outer_instance = self

        class IRQReg(object):
            """IRQ registers"""

            @property
            def rising_edge_control(self):
                """:obj:`int`: The value of all IRQ Control reg, 1 bit represents 1 channel."""
                return i2c_hat.irq.get_reg(Irq.RegName.DI_RISING_EDGE_CONTROL.value)

            @rising_edge_control.setter
            def rising_edge_control(self, value):
                outer_instance._validate_value(value)
                i2c_hat.irq.set_reg(Irq.RegName.DI_RISING_EDGE_CONTROL.value, value)

            @property
            def falling_edge_control(self):
                """:obj:`int`: The value of all IRQ Control reg, 1 bit represents 1 channel."""
                return i2c_hat.irq.get_reg(Irq.RegName.DI_FALLING_EDGE_CONTROL.value)

            @falling_edge_control.setter
            def falling_edge_control(self, value):
                outer_instance._validate_value(value)
                i2c_hat.irq.set_reg(Irq.RegName.DI_FALLING_EDGE_CONTROL.value, value)

            @property
            def capture(self):
                """:obj:`int`: The value of all IRQ Control reg, 1 bit represents 1 channel."""
                return i2c_hat.irq.get_reg(Irq.RegName.DI_CAPTURE.value)

            @capture.setter
            def capture(self, value):
                if value != 0:
                    raise Exception("Value " + str(value) + " not allowed, only 0 is allowed, use 0 to clear the DI IRQ Capture Queue")
                i2c_hat.irq.set_reg(Irq.RegName.DI_CAPTURE.value, value)


        class Channels(object):
            def __getitem__(self, index):
                index = outer_instance._validate_channel_index(index)
                request = outer_instance._i2c_hat._request_frame_(Command.DI_GET_CHANNEL_STATE, [index])
                response = outer_instance._i2c_hat._transfer_(request, 2)
                data = response.data
                if len(data) != 2 or data[0] != index:
                    raise ResponseException('Invalid data')
                return data[1] > 0

            def __len__(self):
                return len(outer_instance.labels)

        class Counters(object):
            def __init__(self, counter_type):
                self.__counter_type = counter_type

            def __getitem__(self, index):
                index = outer_instance._validate_channel_index(index)
                request = outer_instance._i2c_hat._request_frame_(Command.DI_GET_COUNTER, [index, self.__counter_type])
                response = outer_instance._i2c_hat._transfer_(request, 6)
                data = response.data
                if (len(data) != 1 + 1 + 4) or (index != data[0]) or (self.__counter_type != data[1]):
                    raise ResponseException('Invalid data')
                return data[2] + (data[3] << 8) + (data[4] << 16) + (data[5] << 24)

            def __setitem__(self, index, value):
                index = outer_instance._validate_channel_index(index)
                if value != 0:
                    raise ValueError("only '0' is valid, it will reset the counter")
                request = outer_instance._i2c_hat._request_frame_(Command.DI_RESET_COUNTER, [index, self.__counter_type])
                response = outer_instance._i2c_hat._transfer_(request, 2)
                data = response.data
                if (len(data) != 2) or (index != data[0]) or (self.__counter_type != data[1]):
                    raise ResponseException('Invalid data')

            def __len__(self):
                return len(outer_instance.labels)

        self.channels = Channels()
        self.r_counters = Counters(1)
        self.f_counters = Counters(0)
        self.irq_reg = IRQReg()

    @property
    def value(self):
        """:obj:`int`: The value of all the digital inputs, 1 bit represents 1 channel."""
        return self._i2c_hat._get_u32_value_(Command.DI_GET_ALL_CHANNEL_STATES)

    def reset_counters(self):
        """Resets all digital input channel counters of all types(falling and rising edge).

        Raises:
            :obj:`raspihats.i2c_hats._base.ResponseException`: If the response hasn't got the expected format

        """
        request = self._i2c_hat._request_frame_(Command.DI_RESET_ALL_COUNTERS)
        response = self._i2c_hat._transfer_(request, 0)
        data = response.data
        if len(data) != 0:
            raise ResponseException('Invalid data')


class DigitalOutputs(Functionality):
    """Attributes and methods needed for operating the digital outputs channels.

    Args:
        i2c_hat (:obj:`raspihats.i2c_hats._base.I2CHat`): I2CHat instance
        labels (:obj:`list` of :obj:`str`): Labels of digital output channels

    Attributes:
        channels (:obj:`list` of :obj:`bool`): List like object, provides single channel access to digital outputs.

    """

    def __init__(self, i2c_hat, labels):
        Functionality.__init__(self, i2c_hat, labels)
        outer_instance = self

        class Channels(object):
            def __getitem__(self, index):
                index = outer_instance._validate_channel_index(index)
                request = outer_instance._i2c_hat._request_frame_(Command.DQ_GET_CHANNEL_STATE, [index])
                response = outer_instance._i2c_hat._transfer_(request, 2)
                data = response.data
                if len(data) != 2 or data[0] != index:
                    raise ResponseException('unexpected format')
                return data[1] > 0

            def __setitem__(self, index, value):
                index = outer_instance._validate_channel_index(index)
                value = int(value)
                if not (0 <= value <= 1):
                    raise ValueError("'" + str(value) + "' is not a valid value, use: 0 or 1, True or False")
                data = [index, value]
                request = outer_instance._i2c_hat._request_frame_(Command.DQ_SET_CHANNEL_STATE, data)
                response = outer_instance._i2c_hat._transfer_(request, 2)
                if data != response.data:
                    raise ResponseException('unexpected format')

            def __len__(self):
                return len(outer_instance.labels)

        self.channels = Channels()

    @property
    def value(self):
        """:obj:`int`: The value of all the digital outputs, 1 bit represents 1 channel."""
        return self._i2c_hat._get_u32_value_(Command.DQ_GET_ALL_CHANNEL_STATES)

    @value.setter
    def value(self, value):
        self._validate_value(value)
        self._i2c_hat._set_u32_value_(Command.DQ_SET_ALL_CHANNEL_STATES, value)

    @property
    def power_on_value(self):
        """:obj:`int`: Power On Value, this is loaded to outputs at power on."""
        return self._i2c_hat._get_u32_value_(Command.DQ_GET_POWER_ON_VALUE)

    @power_on_value.setter
    def power_on_value(self, value):
        self._validate_value(value)
        self._i2c_hat._set_u32_value_(Command.DQ_SET_POWER_ON_VALUE, value)

    @property
    def safety_value(self):
        """:obj:`int`: Safety Value, this is loaded to outputs at Cwdt Timeout."""

        return self._i2c_hat._get_u32_value_(Command.DQ_GET_SAFETY_VALUE)

    @safety_value.setter
    def safety_value(self, value):
        self._validate_value(value)
        self._i2c_hat._set_u32_value_(Command.DQ_SET_SAFETY_VALUE, value)



"""
This module contains the I2CHat base class.
"""
import sys
import time
import smbus2
import threading
try:
  from enum import Enum
except ImportError:
  from enum34 import Enum
from ._frame import Command, Frame, DecodeException

class ResponseException(Exception):
    """Raised when there's a problem with the I2C-HAT response."""


class I2CHat(object):
    """Implements basic functionality common to all I2C-HATs.

    Args:
        address (int): I2C bus address, valid range depends of base_address
        base_address (int): I2C-HAT family starting address
        board_name (str): I2C-HAT expected board name

    Raises:
        ValueError: If address is not in range

    (*) - attribute value read directly from I2C-HAT

    """
    I2C_PORT = 1 # 1 is default port for Raspberry Pi

    _i2c_bus_lock = threading.Lock()
    _i2c_bus = None

    def __init__(self, address, base_address=None, board_name=None):

        if I2CHat._i2c_bus is None:
            I2CHat._i2c_bus = smbus2.SMBus(I2CHat.I2C_PORT)

        self._address = address
        self._frame_id = 0
        self._transfer_time = None

        if base_address == None:
            if not 0 <= address <= 127:
                raise ValueError("I2C address should be in range[0, 127]")
        else:
            if address & base_address != base_address:
                raise ValueError("I2C address should be in range[" + hex(base_address) + ", " + hex(base_address + 0x0F) + "]")

        if board_name != None:
            if self.name not in board_name:
                raise Exception("unexpected board name '" + self.name + "', expecting '" + board_name + "'")

    def __str__(self):
        string = self.__class__.__name__
        base = 'I2CHat'
        if base not in string:
            string += base
        string += "@" + hex(self._address)
        return string

    def _generate_frame_id_(self):
        """Generate new frame Id, increments current frame Id, wraps to 0x7F beacuse of Raspberry Pi I2C bug which affects MSb

        Returns:
            int: The new frame Id

        """
        self._frame_id += 1
        self._frame_id &= 0x7F
        return self._frame_id

    def _transfer_(self, request_frame, response_data_size, response_expected=True, number_of_tries=5):
        """Tries a number of times to send a request frame and to get a response frame over I2C bus.

        Args:
            request_frame (Frame): Request frame to be sent over the I2C bus
            response_data_size (int): Expected response data size, this is the payload data size
            response_expected (bool): True if a respose is expected
            number_of_tries (int): Number of tries to get the response

        Returns:
            Frame: The response frame

        Raises:
            ResponseException: After all attempts to get a response have failed

        """
        with I2CHat._i2c_bus_lock:
            try_cnt = 1
            while True:
                try:
                    request_data = request_frame.encode()

                    # NOTE: write_i2c_block_data function is used to send commands to the I2C-HAT
                    I2CHat._i2c_bus.write_i2c_block_data(self._address, request_data[0], request_data[1:])

                    if not response_expected:
                        return

                    # NOTE: read_i2c_block_data function sends a i2c_write first, this write has a length of one, and the dummy_byte as payload, this
                    # write will be ignored by the I2C-HAT, after this a i2c_read will be issued, this i2c_read is used for reading the response
                    dummy_byte = 0xFF
                    expected_response_size = Frame.ID_SIZE + Frame.CMD_SIZE + response_data_size + Frame.CRC_SIZE
                    response_data = I2CHat._i2c_bus.read_i2c_block_data(self._address, dummy_byte, expected_response_size)

                    # build response frame
                    response_frame = Frame(request_frame.id, request_frame.cmd)
                    response_frame.decode(response_data)
                    self._transfer_time = time.time()
                    return response_frame

                except (IOError, DecodeException) as ex:
                    if try_cnt >= number_of_tries:
                        if isinstance(ex, IOError):
                            raise ResponseException("no response")
                        else:
                            raise ResponseException(str(ex))
                    time.sleep(0.01)
                    try_cnt += 1

    def _get_u32_value_(self, cmd):
        """Generic get for a unsigned32 value.

        Args:
            cmd (int): Command byte value

        Returns:
            int: The desired unsigned32 value

        Raises:
            ResponseException: If response has bad data length

        """
        request = self._request_frame_(cmd, [])
        response = self._transfer_(request, 4)
        data = response.data
        if len(data) != 4:
            raise ResponseException('invalid response data length')
        return data[0] + (data[1] << 8) + (data[2] << 16) + (data[3] << 24)

    def _set_u32_value_(self, cmd, value):
        """Generic set for a unsigned32 value.

        Args:
            cmd (int): Command byte value
            value (int): Value to be set

        Returns:
            int: The desired unsigned32 value

        Raises:
            ResponseException: I response
        """
        data = [value & 0xFF, (value >> 8) & 0xFF, (value >> 16) & 0xFF, (value >> 24) & 0xFF]
        request = self._request_frame_(cmd, data)
        response = self._transfer_(request, 4)
        if data != response.data:
            raise ResponseException('invalid response data')

    def _request_frame_(self, cmd, data = []):
        """Build request frame, taking care of new frame Id generation.

        Args:
            cmd (int): Frame command byte value
            data (List[int]): Frame payload data

        Returns:
            Frame: The new Frame built with specified parameters

        """
        return Frame(self._generate_frame_id_(), cmd, data)

    @property
    def transfer_time(self):
        """:obj:`float`: Last valid transfer time stamp."""
        return self._transfer_time

    @property
    def address(self):
        """:obj:`int`: I2C bus address."""
        return self._address

    @property
    def name(self):
        """:obj:`string`: Name(*)."""
        request = self._request_frame_(Command.GET_BOARD_NAME)
        response = self._transfer_(request, 25)
        board_name = ''
        for byte in response.data:
            if byte == 0:
                break;
            board_name += chr(byte)
        return board_name

    @property
    def fw_version(self):
        """:obj:`string`: Firmware version(*)."""
        request = self._request_frame_(Command.GET_FIRMWARE_VERSION)
        response = self._transfer_(request, 3)
        data = response.data
        return 'v' + chr(data[0] + 0x30) + '.' + chr(data[1] + 0x30)  + '.' + chr(data[2] + 0x30)

    @property
    def status(self):
        """:obj:`StatusWord`: Status word(*)."""
        return StatusWord(self._get_u32_value_(Command.GET_STATUS_WORD))

    def reset(self):
        """Sends a reset request to the I2C-HAT."""
        request = self._request_frame_(Command.RESET)
        self._transfer_(request, 0, False)

class StatusWord(object):
    """Models StatusWord

    Args:
        value (int): StatusWord raw integer value

    Attributes:
        value (int): StatusWord integer value.
    """

    class Bits(Enum):
        """StatusWord bit masks"""
        POR_RESET = 0x01
        SOFT_RESET = 0x02
        IWD_RESET = 0x04
        CWDT_TIMEOUT = 0x08
        DI_IRQ_CAPTURE_QUEUE_FULL = 0x10

    def __init__(self, value):
        self.value = value

    @property
    def bits(self):
        """:obj:`dict`: StatusWord bit values dictinary."""
        d = dict()
        for bit in self.Bits:
            d[str(bit).replace("Bits.","")] = (self.value & bit.value) != 0x0
        return d

    def __str__(self):
        """Human readable string describing the StatusWord."""
        string = "value: " + hex(self.value) + ", bits: " + str(self.bits)
        return string

class Functionality(object):
    """I2C-HAT functionality base.

    Args:
        i2c_hat (:obj:`raspihats.i2c_hats._base.I2CHat`): I2CHat instance
        labels (:obj:`list` of :obj:`str` or optional): Channel labels

    """

    def __init__(self, i2c_hat, labels=None):
        self._i2c_hat = i2c_hat
        self._labels = labels
        if labels != None:
            self._lc_labels = [l.lower() for l in labels]

    def _validate_channel_index(self, index):
        if self._labels == None:
            raise Exception('no labels defined')

        label = None
        if isinstance(index, int):
            if not (0 <= index < len(self._labels)):
                raise IndexError("'" + str(index) + "' is not a valid channel index")
        elif isinstance(index, str):
            label = index
            try:
                index = self._lc_labels.index(label.lower())
            except ValueError:
                raise ValueError("'" + label + "' is not a valid channel label")
        else:
            raise ValueError("index type is '" + type(index) + "', expecting 'int' or 'str'")
        return index

    def _validate_value(self, value):
        max_value = (0x01 << len(self._labels)) - 1
        if not (0 <= value <= max_value):
            raise ValueError("'" + str(value) + "' is not a valid value, range is [0x00 .. " + hex(max_value) + "]")

    @property
    def labels(self):
        """:obj:`list` of :obj:`str`: Channel Labels."""
        return self._labels

class Cwdt(Functionality):
    """Provides attributes and methods for operating the I2C-HAT CommunicationWatchdogTimer module.

    Args:
        i2c_hat (:obj:`raspihats.i2c_hats._base.I2CHat`): I2CHat instance

    (*) - attribute value read directly from I2C-HAT

    """

    def __init__(self, i2c_hat):
        Functionality.__init__(self, i2c_hat)

    @property
    def period(self):
        """:obj:`float`: The CommunicationWatchdogTimer period value in seconds(*)."""
        return float(self._i2c_hat._get_u32_value_(Command.CWDT_GET_PERIOD)) / 1000

    @period.setter
    def period(self, value):
        if value < 0:
            raise ValueError("period should be greather than zero to enable the CommunicationWatchdogTimer on the I2C-HAT board")
        self._i2c_hat._set_u32_value_(Command.CWDT_SET_PERIOD, int(value * 1000))


class Irq(Functionality):
    """Provides attributes and methods for operating the I2C-HAT IRQ module.

    Args:
        i2c_hat (:obj:`raspihats.i2c_hats._base.I2CHat`): I2CHat instance

    (*) - attribute value read directly from I2C-HAT

    """

    class RegName(Enum):
        """IRQ registers"""
        DI_FALLING_EDGE_CONTROL     = 0x20
        DI_RISING_EDGE_CONTROL      = 0x21
        DI_CAPTURE                  = 0x22

    def __init__(self, i2c_hat):
        Functionality.__init__(self, i2c_hat)

    def get_reg(self, reg_type):
        """:obj:`int`: The value of IRQ control reg, 1 bit represents 1 channel."""
        request = self._i2c_hat._request_frame_(Command.IRQ_GET_REG, [reg_type])
        response = self._i2c_hat._transfer_(request, 5)
        data = response.data
        if (len(data) != 1 + 4) or (data[0] != reg_type):
            raise ResponseException('Invalid data')
        return data[1] + (data[2] << 8) + (data[3] << 16) + (data[4] << 24)

    def set_reg(self, reg_type, value):
        """:obj:`int`: The value of IRQ control reg, 1 bit represents 1 channel."""
        # self._validate_value(value)
        data = [reg_type, value & 0xFF, (value >> 8) & 0xFF, (value >> 16) & 0xFF, (value >> 24) & 0xFF]
        request = self._i2c_hat._request_frame_(Command.IRQ_SET_REG, data)
        response = self._i2c_hat._transfer_(request, 5)
        data = response.data
        if (len(data) != 1 + 4) or (data[0] != reg_type):
            raise ResponseException('Invalid data')
        return data[1] + (data[2] << 8) + (data[3] << 16) + (data[4] << 24)



"""
This module facilitates the raspihats.i2c_hats package integration with robotframework.
All the following functions will be loaded as keywords by robotframework.

Example:

*** Settings ***
Documentation       Test Suite using DI6acDQ6rly I2C-HAT.
Library             raspihats.i2c_hats.robot

*** Test Cases ***

Set Global Variables
    ${i2c_hat_temp}         New DI6acDQ6rly     ${0x60}
    Set Global Variable     ${i2c_hat}          ${i2c_hat_temp}

Board Name
    ${name}                     Get Name    ${i2c_hat}
    Should Be Equal As Strings  ${name}     DI6acDQ6rly I2C-HAT

Read Digital Input
    ${state}    DI Get Channel      ${i2c_hat}      ${0}
    Should Be Equal As Integers     ${state}        ${0}

"""
from . import Di16, Rly10, Di6Rly6, DI16ac, DQ10rly, DQ8rly, DQ16oc, DI6acDQ6rly, DI6acDQ6ssr, DI6dwDQ6ssr

irq_pin = None

def init_irq_pin(pin=21):
    """Initializes the IRQ pin as input with Pull UP enabled. The exported robotframework keyword is 'Init Irq Pin'.

        Args:
            pin (I2CHat): IRQ Pin
    """
    import RPi.GPIO as GPIO
    global irq_pin

    irq_pin = pin
    GPIO.setmode(GPIO.BCM)
    GPIO.setup(irq_pin, GPIO.IN, pull_up_down=GPIO.PUD_UP)

def get_irq_pin_state():
    """Gets the IRQ pin state. The exported robotframework keyword is 'Get Irq Pin State'.

        Returns:
            bool: IRQ pin state
    """
    import RPi.GPIO as GPIO
    global irq_pin

    return GPIO.input(irq_pin)

def exit():
    """Clean Up. The exported robotframework keyword is 'Exit'."""
    import RPi.GPIO as GPIO
    GPIO.cleanup()

def new_Di16(adr):
    """New instance of class Di16. The exported robotframework keyword is 'New Di16'.

        Args:
            adr (int): i2c address

        Returns:
            Di16: A new instance of Di16
    """
    return Di16(adr)

def new_Rly10(adr):
    """New instance of class Rly10. The exported robotframework keyword is 'New Rly10'.

        Args:
            adr (int): i2c address

        Returns:
            Rly10: A new instance of Rly10
    """
    return Rly10(adr)

def new_Di6Rly6(adr):
    """New instance of class Di6Rly6. The exported robotframework keyword is 'New Di6Rly6'.

        Args:
            adr (int): i2c address

        Returns:
            Di6Rly6: A new instance of Di6Rly6
    """
    return Di6Rly6(adr)

def new_DI16ac(adr):
    """New instance of class DI16ac. The exported robotframework keyword is 'New DI16ac'.

        Args:
            adr (int): i2c address

        Returns:
            DI16ac: A new instance of DI16ac
    """
    return DI16ac(adr)

def new_DQ10rly(adr):
    """New instance of class DQ10rly. The exported robotframework keyword is 'New DQ10rly'.

        Args:
            adr (int): i2c address

        Returns:
            DQ10rly: A new instance of DQ10rly
    """
    return DQ10rly(adr)

def new_DQ8rly(adr):
    """New instance of class DQ8rly. The exported robotframework keyword is 'New DQ8rly'.

        Args:
            adr (int): i2c address

        Returns:
            DQ8rly: A new instance of DQ8rly
    """
    return DQ8rly(adr)

def new_DQ16oc(adr):
    """New instance of class DQ16oc. The exported robotframework keyword is 'New DQ16oc'.

        Args:
            adr (int): i2c address

        Returns:
            DQ16oc: A new instance of DQ16oc
    """
    return DQ16oc(adr)

def new_DI6acDQ6rly(adr):
    """New instance of class DI6acDQ6rly. The exported robotframework keyword is 'New DI6acDQ6rly'.

        Args:
            adr (int): i2c address

        Returns:
            DI6acDQ6rly: A new instance of DI6acDQ6rly
    """
    return DI6acDQ6rly(adr)

def new_DI6acDQ6ssr(adr):
    """New instance of class DI6acDQ6ssr. The exported robotframework keyword is 'New DI6acDQ6ssr'.

        Args:
            adr (int): i2c address

        Returns:
            DI6acDQ6ssr: A new instance of DI6acDQ6ssr
    """
    return DI6acDQ6ssr(adr)

def new_DI6dwDQ6ssr(adr):
    """New instance of class DI6dwDQ6ssr. The exported robotframework keyword is 'New DI6dwDQ6ssr'.

        Args:
            adr (int): i2c address

        Returns:
            DI6dwDQ6ssr: A new instance of DI6dwDQ6ssr
    """
    return DI6dwDQ6ssr(adr)

def get_name(i2c_hat):
    """Gets the I2C-HAT name. The exported robotframework keyword is 'Get Name'.

        Args:
            i2c_hat (I2CHat): board

        Returns:
            string: The name of the board
    """
    return i2c_hat.name

def get_firmware_version(i2c_hat):
    """Gets the I2C-HAT name. The exported robotframework keyword is 'Get Firmware Version'.

        Args:
            i2c_hat (I2CHat): board

        Returns:
            string: The firmware version
    """
    return i2c_hat.fw_version

def get_status(i2c_hat):
    """Gets the I2C-HAT status word. The exported robotframework keyword is 'Get Status'.

        Args:
            i2c_hat (I2CHat): board

        Returns:
            int: The status word
    """
    return i2c_hat.status.value

def reset(i2c_hat):
    """Resets the I2C-HAT. The exported robotframework keyword is 'Reset'."""
    i2c_hat.reset()

def cwdt_get_period(i2c_hat):
    """Gets the I2C-HAT CommunicationWatchdogTimer period. The exported robotframework keyword is 'CWDT Get Period'.

        Args:
            i2c_hat (I2CHat): board

        Returns:
            int: The CommunicationWatchdogTimer period
    """
    return i2c_hat.cwdt.period

def cwdt_set_period(i2c_hat, value):
    """Sets the I2C-HAT CommunicationWatchdogTimer period. The exported robotframework keyword is 'CWDT Set Period'.

        Args:
            i2c_hat (I2CHat): board
            value (int): period in seconds
    """
    i2c_hat.cwdt.period = value

def do_get_labels(i2c_hat):
    """Gets the I2C-HAT digital outputs labels. The exported robotframework keyword is 'DO Get Labels'.

        Args:
            i2c_hat (I2CHat): board

        Returns:
            list[sring]: The labels of the digital outputs
    """
    return i2c_hat.dq.labels

def do_get_power_on_value(i2c_hat):
    """Gets the I2C-HAT digital outputs power on value. The exported robotframework keyword is 'DO Get Power On Value'.

        Args:
            i2c_hat (I2CHat): board

        Returns:
            int: The power on value of the digital outputs
    """
    return i2c_hat.dq.power_on_value

def do_set_power_on_value(i2c_hat, value):
    """Sets the I2C-HAT digital outputs power on value. The exported robotframework keyword is 'DO Set Power On Value'.

        Args:
            i2c_hat (I2CHat): board
            value (int): desired power on value
    """
    i2c_hat.dq.power_on_value = value

def do_get_safety_value(i2c_hat):
    """Gets the I2C-HAT digital outputs safety value. The exported robotframework keyword is 'DO Get Safety Value'.

        Args:
            i2c_hat (I2CHat): board

        Returns:
            int: The safety value of the digital outputs
    """
    return i2c_hat.dq.safety_value

def do_set_safety_value(i2c_hat, value):
    """Sets the I2C-HAT digital outputs safety value. The exported robotframework keyword is 'DO Set Safety Value'.

        Args:
            i2c_hat (I2CHat): board
            value (int): desired safety value
    """
    i2c_hat.dq.safety_value = value

def do_get_value(i2c_hat):
    """Gets the I2C-HAT digital outputs value(all channels). The exported robotframework keyword is 'DO Get Value'.

        Args:
            i2c_hat (I2CHat): board

        Returns:
            int: The value(all channels) of the digital outputs
    """
    return i2c_hat.dq.value

def do_set_value(i2c_hat, value):
    """Sets the I2C-HAT digital outputs value(all channels). The exported robotframework keyword is 'DO Set Value'.

        Args:
            i2c_hat (I2CHat): board
            value (int): desired digital outputs value(all channels)
    """
    i2c_hat.dq.value = value

def do_get_channel(i2c_hat, index):
    """Gets the I2C-HAT digital output channel value. The exported robotframework keyword is 'DO Get Channel'.

        Args:
            i2c_hat (I2CHat): board
            index (int): channel index

        Returns:
            boolean: The value of the digital output channel
    """
    return i2c_hat.dq.channels[index]

def do_set_channel(i2c_hat, index, value):
    """Sets the I2C-HAT digital output channel value. The exported robotframework keyword is 'DO Set Channel'.

        Args:
            i2c_hat (I2CHat): board
            index (int): channel index
            value (boolean): desired digital output value

        Returns:
            boolean: The value of the digital output channel
    """
    i2c_hat.dq.channels[index] = value

def di_get_labels(i2c_hat):
    """Gets the I2C-HAT digital inputs labels. The exported robotframework keyword is 'DI Get Labels'.

        Args:
            i2c_hat (I2CHat): board

        Returns:
            list[sring]: The labels of the digital inputs
    """
    return i2c_hat.di.labels

def di_get_value(i2c_hat):
    """Gets the I2C-HAT digital inputs value(all channels). The exported robotframework keyword is 'DI Get Value'.

        Args:
            i2c_hat (I2CHat): board

        Returns:
            int: The value(all channels) of the digital inputs
    """
    return i2c_hat.di.value

def di_get_channel(i2c_hat, index):
    """Gets the I2C-HAT digital input channel value. The exported robotframework keyword is 'DI Get Channel'.

        Args:
            i2c_hat (I2CHat): board
            index (int): channel index

        Returns:
            boolean: The value of the digital input channel
    """
    return i2c_hat.di.channels[index]

def di_get_counter(i2c_hat, index, counter_type):
    """Gets the I2C-HAT digital input channel counter value. The exported robotframework keyword is 'DI Get Counter'.

        Args:
            i2c_hat (I2CHat): board
            index (int): channel index
            counter_type (int): type of counter(0 - falling edge, 1 - rising edge)

        Returns:
            int: The value of the digital input channel counter
    """
    if not (0 <= counter_type <= 1):
        raise ValueError("'" + str(counter_type) + "' is not a valid counter type, use: 0 - falling edge, 1 - rising edge")
    if counter_type == 0:
        value = i2c_hat.di.f_counters[index]
    else:
        value = i2c_hat.di.r_counters[index]
    return value

def di_reset_counter(i2c_hat, index, counter_type):
    """Resets the I2C-HAT digital input channel counter value. The exported robotframework keyword is 'DI Reset Counter'.

        Args:
            i2c_hat (I2CHat): board
            index (int): channel index
            counter_type (int): type of counter(0 - falling edge, 1 - rising edge)
    """
    if not (0 <= counter_type <= 1):
        raise ValueError("'" + str(counter_type) + "' is not a valid counter type, use: 0 - falling edge, 1 - rising edge")
    if counter_type == 0:
        i2c_hat.di.f_counters[index] = 0
    else:
        i2c_hat.di.r_counters[index] = 0

def di_reset_all_counters(i2c_hat):
    """Resets all the I2C-HAT digital input channel counter values. The exported robotframework keyword is 'DI Reset All Counters'.

        Args:
            i2c_hat (I2CHat): board
    """
    i2c_hat.di.reset_counters()

def di_get_irq_reg_rising_edge_control(i2c_hat):
    return i2c_hat.di.irq_reg.rising_edge_control

def di_set_irq_reg_rising_edge_control(i2c_hat, value):
    i2c_hat.di.irq_reg.rising_edge_control = value

def di_get_irq_reg_falling_edge_control(i2c_hat):
    return i2c_hat.di.irq_reg.falling_edge_control

def di_set_irq_reg_falling_edge_control(i2c_hat, value):
    i2c_hat.di.irq_reg.falling_edge_control = value

def di_get_irq_reg_capture(i2c_hat):
    return i2c_hat.di.irq_reg.capture

def di_set_irq_reg_capture(i2c_hat, value):
    i2c_hat.di.irq_reg.capture = value



"""
This module contains the I2C-HATs classes.
"""
from ._base import I2CHat, Cwdt, Irq, ResponseException
from ._digital import DigitalOutputs, DigitalInputs

def set_i2c_port(i2c_port):
    """Set the I2C port number.

    Args:
        i2c_port (int): I2C port number

    """
    import smbus
    I2CHat._i2c_bus = smbus.SMBus(i2c_port)

class Di16(I2CHat):
    """This class exposes all operations supported by the Di16 I2C-HAT.

    Args:
        address (:obj:`int`): I2C bus address, valid range is [0x40, 0x4F]

    Attributes:
        cwdt (:obj:`raspihats.i2c_hats._base.Cwdt`): provides access to CommunicationWatchDogTimer.
        di (:obj:`raspihats.i2c_hats._digital.DigitalInputs`): provides access to DigitalInputs.

    """

    _BASE_ADDRESS = 0x40
    _BOARD_NAME = 'Di16 I2C-HAT'
    _labels = [
        'Di1.1', 'Di1.2', 'Di1.3', 'Di1.4',
        'Di2.1', 'Di2.2', 'Di2.3', 'Di2.4',
        'Di3.1', 'Di3.2', 'Di3.3', 'Di3.4',
        'Di4.1', 'Di4.2', 'Di4.3', 'Di4.4',
    ]

    def __init__(self, address):
        I2CHat.__init__(self, address, self._BASE_ADDRESS, self._BOARD_NAME)
        self.cwdt = Cwdt(self)
        self.di = DigitalInputs(self, self._labels)

class Rly10(I2CHat):
    """This class exposes all operations supported by the Rly10 I2C-HAT.

    Args:
        address (:obj:`int`): I2C bus address, valid range is [0x50, 0x5F]

    Attributes:
        cwdt (:obj:`raspihats.i2c_hats._base.Cwdt`): provides access to CommunicationWatchDogTimer.
        dq (:obj:`raspihats.i2c_hats._digital.DigitalOutputs`): provides access to DigitalOutputs.
    """

    _BASE_ADDRESS = 0x50
    _BOARD_NAME = 'Rly10 I2C-HAT'
    _labels = ['Rly1', 'Rly2', 'Rly3', 'Rly4', 'Rly5', 'Rly6', 'Rly7', 'Rly8', 'Rly9', 'Rly10']

    def __init__(self, address):
        I2CHat.__init__(self, address, self._BASE_ADDRESS, self._BOARD_NAME)
        self.cwdt = Cwdt(self)
        self.dq = DigitalOutputs(self, self._labels)

class Di6Rly6(I2CHat):
    """This class exposes all operations supported by the Di6Rly6 I2C-HAT.

    Args:
        address (:obj:`int`): I2C bus address, valid range is [0x60, 0x6F]

    Attributes:
        cwdt (:obj:`raspihats.i2c_hats._base.Cwdt`): provides access to CommunicationWatchDogTimer.
        di (:obj:`raspihats.i2c_hats._digital.DigitalInputs`): provides access to DigitalInputs.
        dq (:obj:`raspihats.i2c_hats._digital.DigitalOutputs`): provides access to DigitalOutputs.

    """

    _BASE_ADDRESS = 0x60
    _BOARD_NAME = 'Di6Rly6 I2C-HAT'
    _di_labels = ['Di1.1', 'Di1.2', 'Di1.3', 'Di1.4', 'Di1.5', 'Di1.6']
    _dq_labels = ['Rly1', 'Rly2', 'Rly3', 'Rly4', 'Rly5', 'Rly6']

    def __init__(self, address):
        I2CHat.__init__(self, address, self._BASE_ADDRESS, self._BOARD_NAME)
        self.cwdt = Cwdt(self)
        self.di = DigitalInputs(self, self._di_labels)
        self.dq = DigitalOutputs(self, self._dq_labels)

class DI16ac(I2CHat):
    """This class exposes all operations supported by the DI16ac I2C-HAT.

    Args:
        address (:obj:`int`): I2C bus address, valid range is [0x40, 0x4F]

    Attributes:
        cwdt (:obj:`raspihats.i2c_hats._base.Cwdt`): provides access to CommunicationWatchDogTimer.
        di (:obj:`raspihats.i2c_hats._digital.DigitalInputs`): provides access to DigitalInputs.

    """

    _BASE_ADDRESS = 0x40
    _BOARD_NAME = 'DI16ac I2C-HAT'
    _labels = ['I0', 'I1', 'I2', 'I3', 'I4', 'I5', 'I6', 'I7', 'I8', 'I9', 'I10', 'I11', 'I12', 'I13', 'I14', 'I15']


    def __init__(self, address):
        I2CHat.__init__(self, address, self._BASE_ADDRESS, self._BOARD_NAME)
        self.cwdt = Cwdt(self)
        self.irq = Irq(self)
        self.di = DigitalInputs(self, self._labels)

class DQ16oc(I2CHat):
    """This class exposes all operations supported by the DQ16oc I2C-HAT.

    Args:
        address (:obj:`int`): I2C bus address, valid range is [0x50, 0x5F]

    Attributes:
        cwdt (:obj:`raspihats.i2c_hats._base.Cwdt`): provides access to CommunicationWatchDogTimer.
        dq (:obj:`raspihats.i2c_hats._digital.DigitalOutputs`): provides access to DigitalOutputs.
    """

    _BASE_ADDRESS = 0x50
    _BOARD_NAME = 'DQ16oc I2C-HAT'
    _labels = ['Q0', 'Q1', 'Q2', 'Q3', 'Q4', 'Q5', 'Q6', 'Q7', 'Q8', 'Q9', 'Q10', 'Q11', 'Q12', 'Q13', 'Q14', 'Q15']

    def __init__(self, address):
        I2CHat.__init__(self, address, self._BASE_ADDRESS, self._BOARD_NAME)
        self.cwdt = Cwdt(self)
        self.dq = DigitalOutputs(self, self._labels)

class DQ10rly(I2CHat):
    """This class exposes all operations supported by the DQ10rly I2C-HAT.

    Args:
        address (:obj:`int`): I2C bus address, valid range is [0x50, 0x5F]

    Attributes:
        cwdt (:obj:`raspihats.i2c_hats._base.Cwdt`): provides access to CommunicationWatchDogTimer.
        dq (:obj:`raspihats.i2c_hats._digital.DigitalOutputs`): provides access to DigitalOutputs.
    """

    _BASE_ADDRESS = 0x50
    _BOARD_NAME = 'DQ10rly I2C-HAT'
    _labels = ['Q0', 'Q1', 'Q2', 'Q3', 'Q4', 'Q5', 'Q6', 'Q7', 'Q8', 'Q9']

    def __init__(self, address):
        I2CHat.__init__(self, address, self._BASE_ADDRESS, self._BOARD_NAME)
        self.cwdt = Cwdt(self)
        self.dq = DigitalOutputs(self, self._labels)

class DQ8rly(I2CHat):
    """This class exposes all operations supported by the DQ8rly I2C-HAT.

    Args:
        address (:obj:`int`): I2C bus address, valid range is [0x50, 0x5F]

    Attributes:
        cwdt (:obj:`raspihats.i2c_hats._base.Cwdt`): provides access to CommunicationWatchDogTimer.
        dq (:obj:`raspihats.i2c_hats._digital.DigitalOutputs`): provides access to DigitalOutputs.
    """

    _BASE_ADDRESS = 0x50
    _BOARD_NAME = 'DQ8rly I2C-HAT'
    _labels = ['Q0', 'Q1', 'Q2', 'Q3', 'Q4', 'Q5', 'Q6', 'Q7']

    def __init__(self, address):
        I2CHat.__init__(self, address, self._BASE_ADDRESS, self._BOARD_NAME)
        self.cwdt = Cwdt(self)
        self.dq = DigitalOutputs(self, self._labels)

class DI6acDQ6rly(I2CHat):
    """This class exposes all operations supported by the DI6acDQ6rly I2C-HAT.

    Args:
        address (:obj:`int`): I2C bus address, valid range is [0x60, 0x6F]

    Attributes:
        cwdt (:obj:`raspihats.i2c_hats._base.Cwdt`): provides access to CommunicationWatchDogTimer.
        di (:obj:`raspihats.i2c_hats._digital.DigitalInputs`): provides access to DigitalInputs.
        dq (:obj:`raspihats.i2c_hats._digital.DigitalOutputs`): provides access to DigitalOutputs.

    """

    _BASE_ADDRESS = 0x60
    _BOARD_NAME = 'DI6acDQ6rly I2C-HAT'
    _di_labels = ['I0', 'I1', 'I2', 'I3', 'I4', 'I5']
    _dq_labels = ['Q0', 'Q1', 'Q2', 'Q3', 'Q4', 'Q5']

    def __init__(self, address):
        I2CHat.__init__(self, address, self._BASE_ADDRESS, self._BOARD_NAME)
        self.cwdt = Cwdt(self)
        self.irq = Irq(self)
        self.di = DigitalInputs(self, self._di_labels)
        self.dq = DigitalOutputs(self, self._dq_labels)

class DI6acDQ6ssr(I2CHat):
    """This class exposes all operations supported by the DI6acDQ6ssr I2C-HAT.

    Args:
        address (:obj:`int`): I2C bus address, valid range is [0x60, 0x6F]

    Attributes:
        cwdt (:obj:`raspihats.i2c_hats._base.Cwdt`): provides access to CommunicationWatchDogTimer.
        di (:obj:`raspihats.i2c_hats._digital.DigitalInputs`): provides access to DigitalInputs.
        dq (:obj:`raspihats.i2c_hats._digital.DigitalOutputs`): provides access to DigitalOutputs.

    """

    _BASE_ADDRESS = 0x60
    _BOARD_NAME = 'DI6acDQ6ssr I2C-HAT'
    _di_labels = ['I0', 'I1', 'I2', 'I3', 'I4', 'I5']
    _dq_labels = ['Q0', 'Q1', 'Q2', 'Q3', 'Q4', 'Q5']

    def __init__(self, address):
        I2CHat.__init__(self, address, self._BASE_ADDRESS, self._BOARD_NAME)
        self.cwdt = Cwdt(self)
        self.irq = Irq(self)
        self.di = DigitalInputs(self, self._di_labels)
        self.dq = DigitalOutputs(self, self._dq_labels)


class DI6dwDQ6ssr(I2CHat):
    """This class exposes all operations supported by the DI6dwDQ6ssr I2C-HAT.

    Args:
        address (:obj:`int`): I2C bus address, valid range is [0x60, 0x6F]

    Attributes:
        cwdt (:obj:`raspihats.i2c_hats._base.Cwdt`): provides access to CommunicationWatchDogTimer.
        di (:obj:`raspihats.i2c_hats._digital.DigitalInputs`): provides access to DigitalInputs.
        dq (:obj:`raspihats.i2c_hats._digital.DigitalOutputs`): provides access to DigitalOutputs.

    """

    _BASE_ADDRESS = 0x60
    _BOARD_NAME = 'DI6dwDQ6ssr I2C-HAT'
    _di_labels = ['I0', 'I1', 'I2', 'I3', 'I4', 'I5']
    _dq_labels = ['Q0', 'Q1', 'Q2', 'Q3', 'Q4', 'Q5']

    def __init__(self, address):
        I2CHat.__init__(self, address, self._BASE_ADDRESS, self._BOARD_NAME)
        self.cwdt = Cwdt(self)
        self.irq = Irq(self)
        self.di = DigitalInputs(self, self._di_labels)
        self.dq = DigitalOutputs(self, self._dq_labels)


"""
This module contains the I2C Frame class and related classes.
"""
try:
  from enum import Enum
except ImportError:
  from enum34 import Enum
from .. import crc16

class Command(Enum):
    """I2C-HAT commands"""

    # common board commands
    GET_BOARD_NAME              = 0x10
    GET_FIRMWARE_VERSION        = 0x11
    GET_STATUS_WORD             = 0x12
    RESET                       = 0x13

    # Communication WatchDog commands
    CWDT_SET_PERIOD             = 0x14
    CWDT_GET_PERIOD             = 0x15
    IRQ_GET_REG                 = 0x16
    IRQ_SET_REG                 = 0x17

    # Digital Inputs commands
    DI_GET_ALL_CHANNEL_STATES   = 0x20
    DI_GET_CHANNEL_STATE        = 0x21
    DI_GET_COUNTER              = 0x22
    DI_RESET_COUNTER            = 0x23
    DI_RESET_ALL_COUNTERS       = 0x24

    # Digital Outputs commands
    DQ_SET_POWER_ON_VALUE       = 0x30
    DQ_GET_POWER_ON_VALUE       = 0x31
    DQ_SET_SAFETY_VALUE         = 0x32
    DQ_GET_SAFETY_VALUE         = 0x33
    DQ_SET_ALL_CHANNEL_STATES   = 0x34
    DQ_GET_ALL_CHANNEL_STATES   = 0x35
    DQ_SET_CHANNEL_STATE        = 0x36
    DQ_GET_CHANNEL_STATE        = 0x37

class DecodeException(Exception):
    """Raised when I2C Frame decoding fails."""

class Frame(object):
    """The Frame is used for communication over the I2C bus:

    +----+---------+------+------------------------------------------------------------------------------------------+
    | #  | Field   | Size | Description                                                                              |
    +====+=========+======+==========================================================================================+
    | 1. | Id      | 1    | Diferent from one request to another, I2C-HAT responds with the same Id byte.            |
    +----+---------+------+------------------------------------------------------------------------------------------+
    | 2. | Command | 1    | Defines the action to be taken by the I2C-HAT which responds with the same Command byte. |
    +----+---------+------+------------------------------------------------------------------------------------------+
    | 3. | Data    | n    | Payload data.                                                                            |
    +----+---------+------+------------------------------------------------------------------------------------------+
    | 4. | Crc     | 2    | Modbus CRC16 for data integrity.                                                         |
    +----+---------+------+------------------------------------------------------------------------------------------+

    Args:
        id (:obj:`int`): ID byte
        cmd (:obj:`int`): Command byte
        data (:obj:`list` of :obj:`int`): Payload data bytes

    Attributes:
        id (:obj:`int`): ID byte
        cmd (:obj:`int`): Command byte
        data (:obj:`list` of :obj:`int`): Payload data bytes

    """

    # byte size for fields
    ID_SIZE = 1
    CMD_SIZE = 1
    CRC_SIZE = 2

    def __init__(self, id, cmd, data=[]):
        self.id = id
        self.cmd = Command(cmd)
        self.data = data

    def encode(self):
        """Encode the frame fields: Id, Command, Data and Crc to a list of ints.

        Returns:
            :obj:`list` of :obj:`int`: List of frame bytes, raw data that can be transmitted over the I2C bus

        """
        data = [self.id, self.cmd.value] + self.data
        crc = crc16.modbus(data)
        return data + [(crc & 0xFF), ((crc >> 8) & 0xFF)]

    def decode(self, data):
        """Decode raw data from I2C bus. It's used to decode the I2C-HATs response. The fields Id and Command should already be set
        because a valid I2C-HAT response always has the same Id and Command bytes as the request.

        Args:
            data (:obj:`list` of :obj:`int`): Raw I2C data to be decoded

        Raises:
            :obj:`DecodeException`: If the response frame Crc check fails, or has an unexpected Id or Command

        """
        crc = crc16.modbus(data[:-2])
        crc_in = (data[-1] << 8) + data[-2]
        if crc != crc_in:
            #print('crc check failed, ' + hex(crc) + '!=' + hex(crc_in) + str([hex(x) for x in data]))
            raise DecodeException('crc check failed, ' + hex(crc) + '!=' + hex(crc_in) + ' data:' + str([hex(x) for x in data]))
        if self.id != data[0]:
            #print('unexpected id')
            raise DecodeException('unexpected id')
        if self.cmd.value != data[1]:
            #print('unexpected command')
            raise DecodeException('unexpected command')
        self.data = data[2:-2]
